#include <curses.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "error.h"

int myrand(int max)
{
	return (int)(((double)max * (double)rand()) / (double)RAND_MAX);
}

void * mymalloc(int size, char *what)
{
	void *dummy = malloc(size);
	if (!dummy)
		error_exit("failed to allocate %d bytes for %s\n", size, what);

	return dummy;
}

void * myrealloc(void *oldp, int newsize, char *what)
{
	void *dummy = realloc(oldp, newsize);
	if (!dummy)
		error_exit("failed to reallocate to %d bytes for %s\n", newsize, what);

	return dummy;
}

char * mystrdup(char *in)
{
	char *dummy = strdup(in);
	if (!dummy)
		error_exit("failed to copy string '%s': out of memory?\n", in);

	return dummy;
}

void mydelwin(WINDOW *win)
{
	if (ERR == delwin(win))
		error_exit("delwin() failed\n");
}

void mydoupdate(WINDOW *win)
{
	wnoutrefresh(win);
	doupdate();
}

WINDOW * mysubwin(WINDOW *orig, int nlines, int ncols, int begin_y, int begin_x)
{
	WINDOW *dummy = subwin(orig, nlines, ncols, begin_y, begin_x);
	if (!dummy)
		error_exit("failed to create window (subwindow) with dimensions %d-%d at offset %d,%d (terminal size: %d,%d)\n", ncols, nlines, begin_x, begin_y, COLS, LINES);

	return dummy;
}

int resize(void **pnt, int n, int *len, int size)
{
        if (n == *len)
        {
                *len = (*len) ? (*len) * 2 : 4096;
                *pnt = myrealloc(*pnt, (*len) * size, "resize()");
        }
	else if (n > *len || n<0 || *len<0)
		error_exit("resize: fatal memory corruption problem: n > len || n<0 || len<0!\n");

	return 0;
}

ssize_t READ(int fd, char *whereto, size_t len)
{
	ssize_t cnt=0;

	while(len>0)
	{
		ssize_t rc;

		rc = read(fd, whereto, len);

		if (rc == -1)
		{
			if (errno != EINTR)
				error_exit("Problem reading from file\n");
		}
		else if (rc == 0)
		{
			break;
		}
		else
		{
			whereto += rc;
			len -= rc;
			cnt += rc;
		}
	}

	return cnt;
}

char * read_line_fd(int fd)
{
	char *str = mymalloc(80, "read_line_fd buffer");
	int n_in=0, size=80;

	for(;;)
	{
		char c;
		int rc;

		/* read one character */
		if ((rc = READ(fd, &c, 1)) != 1)
		{
			if (n_in == 0)
			{
				free(str);
				str = NULL;
			}
			return str;
		}

		/* resize input-buffer */
		resize((void **)&str, n_in, &size, sizeof(char));

		/* EOF or \n == end of line */
		if (c == 10 || c == EOF)
			break;

		/* ignore CR */
		if (c == 13)
			continue;

		/* add to string */
		str[n_in++] = c;
	}

	/* terminate string */
	resize((void **)&str, n_in, &size, sizeof(char));
	str[n_in] = 0x00;

	return str;
}

/** get_load
 * - in:      nothing
 * - returns: char *
 * this function returns a string containing the current load of the system
 * multitail is running on. format: %1.2f %1.2f %1.2f
 * the string is malloc()ed so the caller should free() things
 */
#if defined(__FreeBSD__) || defined(linux) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__APPLE__) || defined(sun)
char *get_load(void)
{
	double loadavg[3];
	char *str = (char *)mymalloc(20, "loadavg string");

#if defined(__GLIBC__) && ( __GLIBC__ < 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 2))
	/* Older glibc doesn't have getloadavg() - use sysinfo() */
	/* thanks to Ville Herva for this code! */
	double scale = 1 << SI_LOAD_SHIFT;
	struct sysinfo si;

	if (sysinfo(&si) == -1)
	{
		/* let's exit: if these kind of system-
		 * calls start to fail, something must be
		 * really wrong
		 */
		error_exit("sysinfo() failed");
	}

	loadavg[0] = (double)si.loads[0] / scale;
	loadavg[1] = (double)si.loads[1] / scale;
	loadavg[2] = (double)si.loads[2] / scale;
#else
	if (getloadavg(loadavg, 3) == -1)
	{
		/* see comment on sysinfo() */
		error_exit("getloadavg() failed");
	}
#endif
	sprintf(str, "%1.2f %1.2f %1.2f", loadavg[0], loadavg[1], loadavg[2]);

	return str;
}
#endif

/** stop_process
 * - in:      int pid  pid of process
 * - returns: nothing
 * this function sends a TERM-signal to the given process, sleeps for 1009 microseconds
 * and then sends a KILL-signal to the given process if it still exists. the TERM signal
 * is send so the process gets the possibility to gracefully exit. if it doesn't do that
 * in 100 microseconds, it is terminated
 */
void stop_process(int pid)
{
	if (kill(pid, SIGTERM) == -1)
	{
		if (errno != ESRCH)
			error_exit("problem stopping child-process (%d)!\n", pid);
	}

	usleep(1000);

	/* process still exists? */
	if (kill(pid, SIGTERM) == 0)
	{
		/* sleep for a millisecond... */
		usleep(1000);

		/* ...and then really terminate the process */
		if (kill(pid, SIGKILL) == -1)
		{
			if (errno != ESRCH)
				error_exit("problem killing child-process (%d)!\n", pid);
		}
	}
	else if (errno != ESRCH)
		error_exit("problem stopping child-process (%d)!\n", pid);
}

/** delete_array
 * - in:      char **list array of strings to free
 *            int n       number of strings in this array
 * - returns: nothing
 * this function frees an array of strings: all strings are freed and
 * also the pointer-list itself is freed
 */
void delete_array(char **list, int n)
{
	int loop;

	for(loop=n-1; loop>=0; loop--)
		free(list[loop]);

	free(list);
}

int find_char_offset(char *str, char what)
{
	int loop, len = strlen(str);

	for(loop=0; loop<len; loop++)
	{
		if (str[loop] == what)
			return loop;
	}

	return -1;
}

int file_size(char *filename, char exit_error)
{
	struct stat buf;

	if (stat(filename, &buf) == -1)
	{
		if (exit_error == 1)
		{
			error_exit("error %d while accessing file %s\n", errno, filename);
		}
	}

	return buf.st_size;
}
